# (c) Copyright 2011-2012 Cloudera, Inc.

"""This module contains the stat collectors, which when run, will collect stats.
Each stat collector can also report on the location of each stat it collects on
the file system, relative to a root dir."""

import os
import logging
import traceback
from subprocess import Popen, PIPE

logger = logging.getLogger()

class StatResult(object):
    """The result of a stat collection. Encapsulates the result data collected,
    and potentially error_data."""

    def __init__(self, result_data, error_data=None, other=None):
        self.__result_data = result_data
        self.__error_data = error_data
        self.__other = other

    def get_result_data(self):
        return self.__result_data

    def get_error_data(self):
        return self.__error_data

    def get_other(self):
        return self.__other


class Stat(object):
    """The base class for a stat that can be collected. The collect member
    function should be overridden in subclasses."""

    def collect(self, context, dependent_results):
        """Should be overridden by subclasses. Context is a dictionary, which
        Stats have the ability to add to or read from. It is useful if there
        are certain pieces of data a Stat needs that should be fetched while
        collecting the stat. The dependent_results parameter is a dictionary of
        StatResults of other Stats that were collected."""
        raise NotImplementedError


class CommandStat(Stat):
    """A stat that runs a process by launching the process and its arguments
    that are passed into its constructor.
    """

    def __init__(self, command_list, extra_env_path=None):
        """Construct the CommandStat, where the command_list parameter is the
        process and its arguments to launch, and extra_env_path are extra env
        variables to include in the environment of the launching process. The
        first item in the list of command_list is the executable, and the rest
        of the elements are the arguments."""
        self.__command_list = command_list
        if extra_env_path is None:
            extra_env_path = []
        self.__extra_env_path = extra_env_path

    def __str__(self):
        """Returns the command list as a string."""
        return " ".join(self.__command_list)

    def collect(self, context, dependent_results):
        """Collect the standard output and standard error from the command, and
        return those results wrapped in a StatResult instance."""
        try:
            proc = self.__get_command_proc()
            stdout_data, stderr_data = [self.__decode_string(stream) 
                                        for stream in proc.communicate()]
            retcode = proc.wait()
            return StatResult(stdout_data, stderr_data, retcode)
        except OSError, exc:
            command = "".join(self.__command_list)
            logger.error("Error running command %s. %s" % (command, str(exc)))
            raise

    def __decode_string(self, string, encoding="utf-8", errors="replace"):
        """Decodes the string the the given unicode encoding and error replacement
        strategy. You should call this on strings pulled from sysstat commands or
        else non-unicode characters can cause errors when results are written."""
        if string is not None:
            return unicode(string, encoding, errors=errors)
        return None

    def __get_command_proc(self):
        """Runs a command in a process, and returns the Proc instance, which can
        be used to grab the stdout and stderr streams, as well as the return
        code."""
        new_env = os.environ.copy()
        new_env["PATH"] = "%s:/sbin:/usr/sbin" % new_env["PATH"]
        for p in self.__extra_env_path:
            new_env["PATH"] = "%s:%s" % (new_env["PATH"], p)

        logger.info("Running %s" % str(self.__command_list))
        proc = Popen(self.__command_list, stdout=PIPE, stderr=PIPE, stdin=None, env=new_env, close_fds=True)
        return proc

class StatChainRunner(object):
    """Runs a chain of Stat instances. An instance of this should never be
    created directly, but instances should be created using the
    StatChainRunnerBuilder.

    An instance of the builder can be obtained from the static get_builder
    instance of StatChainRunner. The builder can set the persister, and the
    Stat instances to run, as well as determine the dependency among stats,
    and verify that they be run in the right order. An example of using the
    builder:

    my_stat_chain_runner = StatChainRunner.get_builder(). \
        set_persister(some_persister). \
        add("stat_key1", some_stat_instance1). \
        add("stat_key2", some_stat_instance2, ["stat_key1"]). \
        build()
    """

    class StatChainRunnerBuilderException(Exception):
        """An exception type for exceptions that occur during construction
        of a StatChainRunner."""
        pass

    class StatChainRunnerBuilder(object):
        """A builder that is responsible for constructing StatChainRunner
        instances. Stat instances can be added using the add method, and
        a persister can be set using the set_persister method.

        When build is called, the builder determines the order the Stat
        instances should be run in, verifies that there are no cyclical
        dependencies, and verifies that a persister was provided. If there
        are errors, build will throw StatChainRunnerBuilderException,
        otherwise, it will pass that info along to a StatChainRunner
        instance that can be used to run and collect the stats.
        """

        def __init__(self):
            self.__persister = None
            self.__chain = {}

        def set_persister(self, persister):
            """Sets a persister that will be called for each stat collected.
            Returns self so method calls can be chained."""
            self.__persister = persister
            return self

        def remove_persister(self):
            """Removes a persister from the builder. Returns self so method
            calls can be chained."""
            self.__persister = None
            return self

        def add(self, key, stat, dependency_keys=None):
            """Adds a stat to the chain. Each stat must have a key so that it
            can be referenced as a dependency. The dependency_keys parameter
            is a list of keys this stat has a dependency on, which means those
            dependent stats will run before this stat, and their results will
            be passed to this stat.

            Returns self so that method calls can be chained.
            """
            if dependency_keys is None:
                dependency_keys = []
            if key in self.__chain:
                msg = "Key %s already in chain." % (key,)
                raise StatChainRunner.StatChainRunnerBuilderException(msg)
            self.__chain[key] = (stat, dependency_keys)
            return self

        def add_all(self, stat_tuples):
            """Given a list of stat_tuples, which are of the form (key, stat,
            dependency_keys) (the same parameters sent to the add member func,
            it will add all those stat_tuples to the chain."""
            for key, stat, dependency_keys in stat_tuples:
                self.add(key, stat, dependency_keys)

        def remove(self, key):
            """Removes stat with key represented by parameter key from the chain
            Returns self so that method calls can be chained."""
            del self.__chain[key]
            return self

        def clear(self):
            """Clears all stats from the chain. Returns self so that method
            calls can be chained."""
            self.__chain.clear()
            return self

        def get_stats_dict(self):
            stats_dict = {}
            for stat_key, (stat, _) in self.__chain.iteritems():
                stats_dict[stat_key] = stat
            return stats_dict

        def build(self):
            """Will build a StatChainRunner instance from the Stat instances
            that should be run, and the persister. This will throw an exception
            if a persister has not been provided. It will throw an exception if
            there are cycles in the dependency chain. Finally, it will throw an
            exception if a dependency was specified that does not exist.

            This function will return the StatChainRunner instance that is
            configured to run the Stat instances in a valid order.
            """
            if self.__persister is None:
                msg = "Must provide a persister to build a StatChainRunner"
                raise StatChainRunner.StatChainRunnerBuilderException(msg)
            run_order = self.__get_run_order()
            return StatChainRunner(self.__chain, run_order, self.__persister)

        def __get_run_order(self):
            run_order_keys = []
            for stat_key, (stat, dependency_keys) in self.__chain.iteritems():
                if len(dependency_keys) == 0 and stat_key not in run_order_keys:
                    run_order_keys.insert(0, stat_key)
                elif stat_key not in run_order_keys:
                    if self.__are_there_cycles(set([stat_key]), dependency_keys):
                        msg = "Cycle in the stat dependencies: %s" % (str(self.__chain))
                        raise StatChainRunner.StatChainRunnerBuilderException(msg)
                    self.__append_dependencies(run_order_keys, dependency_keys)
                    run_order_keys.append(stat_key)
                    
            return run_order_keys

        def __append_dependencies(self, dep_list, dependency_keys):
            for dependency_key in dependency_keys:
                if dependency_key not in self.__chain:
                    msg = "Dependency key %s not in chain." % (dependency_key)
                    raise StatChainRunner.StatChainRunnerBuilderException(msg)
                self.__append_dependencies(dep_list, self.__chain[dependency_key][1])
                if dependency_key not in dep_list:
                    dep_list.append(dependency_key)

        def __are_there_cycles(self, ancestor_keys, dependency_keys):
            for ancestor_key in ancestor_keys:
                if ancestor_key in dependency_keys:
                    return True
            for dependency_key in dependency_keys:
                if dependency_key not in self.__chain:
                    msg = "Dependency key %s not in chain." % (dependency_key)
                    raise StatChainRunner.StatChainRunnerBuilderException(msg)
                new_ancestor_keys = set(ancestor_keys)
                new_ancestor_keys.add(dependency_key)
                if self.__are_there_cycles(
                    new_ancestor_keys, self.__chain[dependency_key][1]):
                    return True
            return False

    @staticmethod
    def get_builder():
        """Get a builder instance that knows how to construct a StatChainRunner
        instance."""
        return StatChainRunner.StatChainRunnerBuilder()

    def __init__(self, chain, run_order, persister):
        """This constructor should never be called by user code, but should be
        called by a builder."""
        self.__chain = chain
        self.__run_order = run_order
        self.__persister = persister
        self.__dependency_keys = set()
        for _, dependencies in chain.itervalues():
            for dependency in dependencies:
                self.__dependency_keys.add(dependency)

    def run(self, context=None):
        """Runs the chain of Stat instances. The results of any dependencies
        will be passed to any stat which has a dependency on those. After
        each stat collect is run, the results will be sent to the configured
        persister. Context can be a dictionary that will be passed to each
        Stat, which can read and modify it, when that Stat's collect method
        is invoked."""
        dependency_results = {}
        if context is None:
            context = {}
        for stat_key in self.__run_order:
            stat, dependencies = self.__chain[stat_key]
            dependent_results = {}
            for dependency in dependencies:
                dependent_results[dependency] = dependency_results[dependency]
            try:
                stat_result = stat.collect(context, dependent_results)
            except Exception, exc:
                logger.error("Error collecting stat: %s" % (str(exc),))
                logger.error(traceback.format_exc())
                continue
            if stat_key in self.__dependency_keys:
                dependency_results[stat_key] = stat_result
            self.__persister(stat_key, stat_result)

    def to_builder(self):
        """Constructs a builder initialized with the current StatChainRunner.
        This builder can then be modified in any way to construct a new
        StatChainRunner. This is very usefule when testing, for example, if
        you want to take a StatChainRunner, and stub out its persister, or
        remove or modify any stats to be run.
        """
        builder = StatChainRunner.StatChainRunnerBuilder()
        builder.set_persister(self.__persister)
        for stat_key, stat_deps_pair in self.__chain.iteritems():
            stat = stat_deps_pair[0]
            dependency_keys = stat_deps_pair[1]
            builder.add(stat_key, stat, dependency_keys)
        return builder

    def get_run_order(self):
        """Return the run order of the keys."""
        # Return a clone so the run order can't be modified
        return list(self.__run_order)

    def get_stats_dict(self):
        return self.to_builder().get_stats_dict()

# NOTE:  169.254.169.254 is the special link-local address that Amazon uses to
# pass data into an EC2 instance. Should not be bound to anything in any other 
# normal domain.
EC2_PROBE_CMD = "curl -m 1 http://169.254.169.254/2011-01-01/meta-data/instance-type"

def create_sys_stat_runner(machine_collection_dao=None):
    """Returns a system stats runner that collects the output of various commands"""
    def persist_sys_stat(stat_key, stat_result):
        if machine_collection_dao is None:
            return
        machine_collection_dao.put_sys_stat(
            stat_key, stat_result.get_result_data(), 
            stat_result.get_error_data(), stat_result.get_other())
        machine_collection_dao.update_performance(stat_key)

    return StatChainRunner.get_builder(). \
        set_persister(persist_sys_stat). \
        add("df",              CommandStat(["df", "-k"])). \
        add("cpuinfo",         CommandStat(["cat", "/proc/cpuinfo"])). \
        add("meminfo",         CommandStat(["cat", "/proc/meminfo"])). \
        add("interrupts",      CommandStat(["cat", "/proc/interrupts"])). \
        add("mounts",          CommandStat(["cat", "/proc/mounts"])). \
        add("swaps",           CommandStat(["cat", "/proc/swaps"])). \
        add("diskstats",       CommandStat(["cat", "/proc/diskstats"])). \
        add("procvmstat",      CommandStat(["cat", "/proc/vmstat"])). \
        add("uptime",          CommandStat(["cat", "/proc/uptime"])). \
        add("security_limits", CommandStat(["cat", "/etc/security/limits.conf"])). \
        add("ifconfig",        CommandStat(["ifconfig", "-a"])). \
        add("ethtool",         CommandStat(["ethtool", "eth0"])). \
        add("ethtool-s",       CommandStat(["ethtool", "-S", "eth0"])). \
        add("lsmod",           CommandStat(["lsmod"])). \
        add("chkconfig",       CommandStat(["chkconfig", "--list"])). \
        add("lspci",           CommandStat(["lspci"])). \
        add("lscpu",           CommandStat(["lscpu"])). \
        add("lshw",            CommandStat(["lshw"])). \
        add("lsb_release",     CommandStat(["lsb_release", "-a"])). \
        add("java_version",    CommandStat(["java", "-version"])). \
        add("uname-a",         CommandStat(["uname", "-a"])). \
        add("uname-r",         CommandStat(["uname", "-r"])). \
        add("issue",           CommandStat(["cat", "/etc/issue"])). \
        add("redhat-release",  CommandStat(["cat", "/etc/redhat-release"])). \
        add("hosts",           CommandStat(["cat", "/etc/hosts"])). \
        add("resolv",          CommandStat(["cat", "/etc/resolv.conf"])). \
        add("nsswitch",        CommandStat(["cat", "/etc/nsswitch.conf"])). \
        add("sysctl",          CommandStat(["/sbin/sysctl", "-A"])). \
        add("service",         CommandStat(["service",  "--status-all"])). \
        add("dmidecode",       CommandStat(["dmidecode"])). \
        add("rpm-qa",          CommandStat(["rpm", "-qa"])). \
        add("dpkg-l",          CommandStat(["dpkg", "-l"])). \
        add("vmstat",          CommandStat(["vmstat"])). \
        add("ec2inst_type",    CommandStat(EC2_PROBE_CMD.split(" "))). \
        add("lsof",            CommandStat(["lsof", "-i", "-n", "-P"])). \
        add("ps_aux",          CommandStat(["ps", "aux"])). \
        add("host_conf",       CommandStat(["cat", "/etc/host.conf"])). \
        add("yum_repos",       CommandStat(["ls", "/etc/yum.repos.d"])). \
        add("apt_sources",     CommandStat(["cat", "/etc/apt/sources.list"])). \
        add("zypper_repos",    CommandStat(["zypper", "repos", "-d"])). \
        add("lvdisplay",       CommandStat(["lvdisplay"])). \
        add("dmesg",           CommandStat(["dmesg"])). \
        add("selinux",         CommandStat(["cat", "/etc/sysconfig/selinux"])). \
        add("suse-release",    CommandStat(["cat", "/etc/suse-release"])). \
        add("sar",             CommandStat(["sar", "-A"])). \
        add("swappiness",      CommandStat(["cat", "/proc/sys/vm/swappiness"])). \
        add("iptables",        CommandStat(["iptables", "-L", "-v", "-n"])). \
        add("transparent_hue_page", CommandStat(["cat", "/sys/kernel/mm/redhat_transparent_hugepage/enabled"])). \
        add("top",             CommandStat(["top", "-b", "-n 1"])). \
        add("ifcfg-rhel",      CommandStat(["cat", "/etc/sysconfig/network-scripts/ifcfg-eth0"])). \
        add("ifcfg-sles",      CommandStat(["cat", "/etc/sysconfig/network/ifcfg-eth0"])). \
        add("network",         CommandStat(["cat", "/etc/sysconfig/network"])). \
        add("etc_hostname",    CommandStat(["cat", "/etc/hostname"])). \
        add("hostname",        CommandStat(["hostname", "--fqdn"])). \
        build()
