# Copyright (c) 2012 Cloudera, Inc. All rights reserved.

import logging
import time

from cmf.monitor.abstract_monitor import AbstractMonitor
from cmf.monitor.monitor_properties import ConfParser
from cmf.monitor.base import DebugStatusReport
from cmf.monitor.proc_metrics_utils import add_proc_metrics
from cmf.monitor.role_update import RoleUpdate

LOG = logging.getLogger('HbaseThriftServerMonitor')

_ROLETYPE = 'HBASETHRIFTSERVER'

# Section and key names in cloudera-monitor.properties
# Since CM 4.5
_SECTION = 'hbasethriftserver'
_SERVICE_NAME = 'service_name'
_ROLE_NAME = 'role_name'
_HOST = 'host'
_LOG_DIR = 'log_dir'
_COLLECT_INTERVAL = 'collect_interval'
_SERVICE_VERSION = 'service_version'

class HbaseThriftServerMonitor(AbstractMonitor):

  def __init__(self, daemon, registry):
    AbstractMonitor.__init__(self, daemon, registry)
    self._last_conf_path = None
    self._last_reported_timestamp = None
    self._last_collection_ended = None
    self._last_collection_length = None
    self._role_dir_monitors = [ ]
    self._metric_accessors = daemon.metric_accessors

    # Config data
    self._name = None
    self._service_name = None
    self._service_version = None
    self._log_dir = None

  def __str__(self):
    return "HbaseThriftServerMonitor for %s" % (self._name,)

  def _reload_config(self, confparse):
    """Returns True/False for success"""
    try:
      # Extraction
      name = confparse.get(_SECTION, _ROLE_NAME)
      service = confparse.get(_SECTION, _SERVICE_NAME)
      service_version = \
          confparse.getint_with_default(_SECTION, _SERVICE_VERSION, 3)
      log_dir = confparse.get_with_default(_SECTION, _LOG_DIR, None)
      host = confparse.get(_SECTION, _HOST)
    except ValueError, ex:
      self.record_error(ex)
      LOG.error("Value error in %s: %s" % (confparse.path, ex))
      return False

    # Assign
    self._name = name
    self._service_name = service
    self._service_version = service_version
    self._log_dir = log_dir

    # Make monitors for the log directory and the data directories.
    self._role_dir_monitors = [ ]
    if self._log_dir is not None:
      self.create_dir_monitor(self._log_dir, self._role_dir_monitors)

    return True

  def collect(self):
    """Overrides AbstractMonitor. Collect HbaseThriftServer metrics. Invoked
       asynchronously."""
    start = time.time()

    # Reload config if necessary
    confparse = self.conf_parser
    if self._last_conf_path != confparse.path:
      if not self._reload_config(confparse):
        return
      self._last_conf_path = confparse.path

    # Collect directory info
    for mon in self._role_dir_monitors:
      mon.collect()

    # If our pid value is None, it means that the associated process is not
    # running. If this is the case, we skip metric collection.
    if self._pid is None:
      return

    self._last_collection_ended = time.time()
    self._last_collection_length = self._last_collection_ended - start

  def parse_metrics(self, json):
    pass

  def get_role_update(self):
    """Returns a RoleUpdate to be sent to the firehose"""
    update = RoleUpdate(name=self._name,
                        roletype=_ROLETYPE,
                        service_name=self._service_name)

    add_proc_metrics(update, self._pid, LOG)

    for mon in self._role_dir_monitors:
      update.add_dir_update(mon.get_role_dir_update())

    return update

  def get_debug_status_report(self):
    return DebugStatusReport(config_items=self.conf_parser.items(_SECTION),
                             errors=(self.last_error,),
                             last_collection_ended=self._last_collection_ended,
                             last_collection_length=self._last_collection_length)

  def get_subject_type(self):
    return _ROLETYPE
