// (c) Copyright 2011-2012 Cloudera, Inc. All rights reserved.
define([
  'cloudera/cmf/charts/Plot',
  'cloudera/cmf/charts/TimeSeriesVisualizer',
  'underscore'
], function(Plot, TimeSeriesVisualizer, _) {
  describe("TimeSeriesVisualizer Tests", function() {
    var module, id="tcVisualizerContainer", options = {
      plot: {
        tsquery: "Some Query",
        chartType: Plot.ChartType.LINE
      },
      container: "#" + id,
      enableFacets: true
    }, mockVisualizer = {
      render: function() {}
    };

    beforeEach(function() {
      $('<div id="' + id + '"></div>').appendTo(document.body);
    });

    afterEach(function() {
      module.unsubscribe();
      $("#" + id).remove();
    });

    it("should create a TimeSeriesVisualizer object", function() {
      module = new TimeSeriesVisualizer(options);

      expect(module.facets).toBeDefined();
      expect(module.getViewModel()).toBeDefined();
    });

    it("should test the TimeSeriesVisualizer.render method", function() {
      module = new TimeSeriesVisualizer(options);

      spyOn(module.facets, 'update');
      spyOn(module, 'createVisualizer').andReturn(mockVisualizer);
      spyOn(mockVisualizer, 'render');

      var timeSeriesResponse = {
        timeSeries: [ {
          data: [{x:1,y:2}],
          metadata: {
            attributes: {}
          }
        }, {
          data: [{x:2,y:3}],
          metadata: {
            attributes: {}
          }
        }, {
          // should ignore this one because it is empty
          data: [],
          metadata: {}
        }, {
          // should ignore this one too.
        } ]
      };
      module.render(timeSeriesResponse);

      expect($("#" + id).find(".yAxis").length).toEqual(2);
      expect($("#" + id).find(".chart").length).toEqual(2);
      expect($("#" + id).find(".chart-container").length).toEqual(2);
      expect($("#" + id).find(".chart-title").length).toEqual(3);

      expect(mockVisualizer.render).wasCalled();
      expect(module.vizs.length).toEqual(2);
      expect(module.facets.update).wasCalled();
    });

    it("should test the TimeSeriesVisualizer.render method without yAxis", function() {
      var newOptions = $.extend(true, {}, options);
      newOptions.plot.showYAxis = false;
      module = new TimeSeriesVisualizer(newOptions);

      spyOn(module.facets, 'update');
      spyOn(module, 'createVisualizer').andReturn(mockVisualizer);
      spyOn(mockVisualizer, 'render');

      var timeSeriesResponse = {
        timeSeries: [ {
          data: [{x:1,y:2}],
          metadata: {
            attributes: {}
          }
        }, {
          data: [{x:2,y:3}],
          metadata: {
            attributes: {}
          }
        }, {
          // should ignore this one because it is empty
          data: [],
          metadata: {}
        }, {
          // should ignore this one too.
        } ]
      };
      module.render(timeSeriesResponse);

      expect($("#" + id).find(".yAxis").length).toEqual(0);
      expect($("#" + id).find(".chart").length).toEqual(2);
      expect($("#" + id).find(".chart-container").length).toEqual(2);
      expect($("#" + id).find(".chart-title").length).toEqual(3);

      expect(mockVisualizer.render).wasCalled();
      expect(module.vizs.length).toEqual(2);
      expect(module.facets.update).wasCalled();
    });

    it("should test the TimeSeriesVisualizer.regroupTimeSeries method", function() {
      module = new TimeSeriesVisualizer(options);

      var timeSeriesArray = [ {
        data: [{x:1,y:2}],
        metadata: {
          attributes: {
            "a": "a0",
            "b": "b0"
          }
        }
      }, {
        data: [{x:2,y:3}],
        metadata: {
          attributes: {
            "a": "a0",
            "b": "b1"
          }
        }
      }, {
        data: [{x:3,y:4}],
        metadata: {
          attributes: {
            "a": "a0",
            "b": "b1"
          }
        }
      } ];

      var grouped = module.regroupTimeSeries(timeSeriesArray, Plot.FACETTING_SINGLE_PLOT);
      expect(grouped.all).toEqual(timeSeriesArray);

      grouped = module.regroupTimeSeries(timeSeriesArray, "a");
      expect(grouped.a0.length).toEqual(3);

      grouped = module.regroupTimeSeries(timeSeriesArray, "b");
      expect(grouped.b0.length).toEqual(1);
      expect(grouped.b1.length).toEqual(2);

      grouped = module.regroupTimeSeries(timeSeriesArray, "a,b");
      expect(grouped["a0,b0"].length).toEqual(1);
      expect(grouped["a0,b1"].length).toEqual(2);

      grouped = module.regroupTimeSeries(timeSeriesArray, Plot.FACETTING_NONE);
      expect(grouped[0][0]).toEqual(timeSeriesArray[0]);
      expect(grouped[1][0]).toEqual(timeSeriesArray[1]);
      expect(grouped[2][0]).toEqual(timeSeriesArray[2]);
    });

    it("should catch the clickFacetGroup event", function() {
      module = new TimeSeriesVisualizer(options);

      spyOn(module, "render");
      $.publish("clickFacetGroup", [Plot.FACETTING_NONE]);
      expect(module.render).wasCalled();
    });

    it("should test the various actions", function() {
      var newOptions = $.extend(true, {
        enableEditing: true,
        enableCloning: true,
        enableRemoving: true
      }, options);
      module = new TimeSeriesVisualizer(newOptions);

      spyOn(module, 'createVisualizer').andReturn(mockVisualizer);
      spyOn(mockVisualizer, 'render');

      var timeSeriesResponse = {
        timeSeries: [ {
          data: [{x:1,y:2}],
          metadata: {
            attributes: {}
          }
        }, {
          data: [{x:2,y:3}],
          metadata: {
            attributes: {}
          }
        }, {
          // should ignore this one because it is empty
          data: [],
          metadata: {}
        }, {
          // should ignore this one too.
        } ]
      };
      module.render(timeSeriesResponse);
      spyOn($, "publish");

      var $container = $("#" + id);
      $container.find(".edit-chart").trigger("click");
      expect($.publish).wasCalled();
      expect($.publish.mostRecentCall.args[0]).toEqual("editPlot");

      $container.find(".clone-chart").trigger("click");
      expect($.publish).wasCalled();
      expect($.publish.mostRecentCall.args[0]).toEqual("clonePlot");

      $container.find(".remove-chart").trigger("click");
      expect($.publish).wasCalled();
      expect($.publish.mostRecentCall.args[0]).toEqual("removePlot");
    });

    it("should test adjustWidth", function() {
      module = new TimeSeriesVisualizer(options);

      spyOn(module, "adjustTitleWidth");
      spyOn(module, "adjustChartWidth");

      module.adjustWidth(200);

      expect(module.adjustTitleWidth.mostRecentCall.args[1]).toBeDefined();
      expect(module.adjustChartWidth.mostRecentCall.args[1]).toBeDefined();
    });

    it('should render a legend conditional on the number of streams', function() {
      var createFakePlot = function() {
        return {
          data: [{x: 1, y: 2}],
          metadata: {
            attributes: {}
          }
        };
      };
      var newOptions = _.defaults({
        plot: {
          tsquery: "Some Query",
          chartType: Plot.ChartType.LINE,
          showLegend: true,
          facetting: Plot.FACETTING_SINGLE_PLOT
        }
      }, options);
      module = new TimeSeriesVisualizer(newOptions);
      var timeSeriesResponse = {
        timeSeries: _.map(_.range(50), createFakePlot)
      };
      spyOn(module, 'addLegendContainer');
      spyOn(module, 'createVisualizer').andCallThrough();
      module.render(timeSeriesResponse);
      expect(module.addLegendContainer).wasCalled();
      module.addLegendContainer.reset();
      var visualizerArgs = module.createVisualizer.mostRecentCall.args;
      // Verify that we told the visualizer we want a legend.
      expect(visualizerArgs[0].showLegend).toBeTruthy();

      // Add another streams and we're over the threshold.
      timeSeriesResponse.timeSeries.push(createFakePlot());
      module.render(timeSeriesResponse);
      expect(module.addLegendContainer).wasNotCalled();
      // Verify that we told the visualizer we don't want a legend.
      visualizerArgs = module.createVisualizer.mostRecentCall.args;
      expect(visualizerArgs[0].showLegend).toBeFalsy();
    });
  });
});
